/*
 * Copyright (c) 2002-2004 Atheros Communications, Inc.
 * All rights reserved.
 *
 * $Id: //depot/sw/branches/linuxsrc_5.0.3/src/802_11/madwifi/tx99/ath_tx99.c#1 $
 */

/*
 * ath_tx99: continuous transmit diagnostic utility.
 *
 * To use, build the module.  Load the driver as usual followed by
 * the ath_tx99 module.  On module load specify the base device to
 * use for testing; e.g.
 *	modprobe ath_tx99 device=wifi0
 *
 * To start continuous transmit bring the device up; e.g.
 *	ifconfig wifi0 up
 * To stop continuous transmit mark the device down:
 *	ifconfig wifi0 down
 * or unload the ath_tx99 module.  Note that so long as the
 * module is loaded you can only do continuous transmit.
 *
 * By default the current channel is used.  To specify a channel
 * for transmits use sysctl; e.g.
 *	sysctl -w dev.wifi0.tx99freq=5220
 * The frequency is used to find a channel based on the current
 * band/wireless mode.  This can be overridden with:
 *	sysctl -w dev.wifi0.tx99mode=X
 * (where X is the value of IEEE80211_MODE_*; e.g. 3 for 11g).
 * Finally the transmit rate can be specified in Kbps; e.g.
 *	sysctl -w dev.wifi0.txrate=54000
 * and the transmit power in dBm with
 *	sysctl -w dev.wifi0.txpower=30
 * Note that you must enable tpc for the transmit power setting
 * to have an effect as it is set in the tx descriptor.
 */
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/random.h>
#include <linux/delay.h>
#include <linux/cache.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>

#include <asm/uaccess.h>

#include "if_ethersubr.h"		/* for ETHER_IS_MULTICAST */
#include "if_media.h"
#include "if_llc.h"
#include "net80211/if_athproto.h"
#include <net80211/ieee80211_var.h>

#include "if_athvar.h"
#include "ah_desc.h"
#include "ah_devid.h"

#ifdef ATH_PCI		/* PCI BUS */
#include "if_ath_pci.h"
#endif			/* PCI BUS */
#ifdef ATH_AHB		/* AHB BUS */
#include "if_ath_ahb.h"
#endif			/* AHB BUS */
#include "ath_tx99.h"

#include "ah.h"
#include "ar5212/ar5212reg.h"
#include "ar5212/ar5212phy.h"

#define	OS_REG_SET_BIT(_a, _r, _f) \
	OS_REG_WRITE(_a, _r, OS_REG_READ(_a, _r) | (_f))
#define	OS_REG_CLR_BIT(_a, _r, _f) \
	OS_REG_WRITE(_a, _r, OS_REG_READ(_a, _r) &~ (_f))
#define	SM(_v, _f)	(((_v) << _f##_S) & _f)
#define	MS(_v, _f)	(((_v) & _f) >> _f##_S)

static u_int8_t PN9Data[] = {0xff, 0x87, 0xb8, 0x59, 0xb7, 0xa1, 0xcc, 0x24, 
			     0x57, 0x5e, 0x4b, 0x9c, 0x0e, 0xe9, 0xea, 0x50, 
			     0x2a, 0xbe, 0xb4, 0x1b, 0xb6, 0xb0, 0x5d, 0xf1, 
			     0xe6, 0x9a, 0xe3, 0x45, 0xfd, 0x2c, 0x53, 0x18, 
			     0x0c, 0xca, 0xc9, 0xfb, 0x49, 0x37, 0xe5, 0xa8, 
			     0x51, 0x3b, 0x2f, 0x61, 0xaa, 0x72, 0x18, 0x84, 
			     0x02, 0x23, 0x23, 0xab, 0x63, 0x89, 0x51, 0xb3, 
			     0xe7, 0x8b, 0x72, 0x90, 0x4c, 0xe8, 0xfb, 0xc0};
static u_int8_t test_addr[6] ={ 0x0, 0x0, 0x2, 0x3, 0x0, 0x0};

struct ath_tx99_priv {
	struct ath_tx99 base;		/* driver-visible state */

	struct net_device *dev;
	u_int32_t	mode;		/* experimental mode */
	u_int32_t	txpower;	/* tx power */
	u_int32_t	txrate;		/* tx rate */
	u_int32_t	txfreq;		/* tx channel frequency */
	u_int32_t	txmode;		/* tx channel mode, e.g. 11b */

	struct sk_buff		*skb;
	STAILQ_HEAD(tpc_buf,ath_buf) tmp_q;

	struct ctl_table_header	*sysctl_header;
	struct ctl_table	*sysctls;
};

static void
ath_tx_draintxq(struct ath_softc *sc, struct ath_txq *txq)
{
	struct ath_buf *bf;

	/*
	 * NB: this assumes output has been stopped and
	 *     we do not need to block ath_tx_tasklet
	 */
	for (;;) {
		ATH_TXQ_LOCK(txq);
		bf = STAILQ_FIRST(&txq->axq_q);
		if (bf == NULL) {
			txq->axq_link = NULL;
			txq->axq_linkbuf = NULL;
			ATH_TXQ_UNLOCK(txq);
			break;
		}
		ATH_TXQ_REMOVE_HEAD(txq, bf_list);
		ATH_TXQ_UNLOCK(txq);
		bus_unmap_single(sc->sc_bdev,
			bf->bf_skbaddr, bf->bf_skb->len, BUS_DMA_TODEVICE);
		dev_kfree_skb(bf->bf_skb);
		bf->bf_skb = NULL;
		if (bf->bf_node != NULL) {
			ieee80211_free_node(bf->bf_node);
			bf->bf_node = NULL;
		}

		ATH_TXBUF_LOCK(sc);
		STAILQ_INSERT_TAIL(&sc->sc_txbuf, bf, bf_list);
		ATH_TXBUF_UNLOCK(sc);
	}
}

static void
ath_tx_stopdma(struct ath_softc *sc, struct ath_txq *txq)
{
	struct ath_hal *ah = sc->sc_ah;

	(void) ath_hal_stoptxdma(ah, txq->axq_qnum);
}

/*
 * Drain the transmit queues and reclaim resources.
 */
static void
ath_draintxq(struct ath_softc *sc)
{
	struct ath_hal *ah = sc->sc_ah;
	int i;

	/* XXX return value */
	if (!sc->sc_invalid) {
		(void) ath_hal_stoptxdma(ah, sc->sc_bhalq);
		for (i = 0; i < HAL_NUM_TX_QUEUES; i++)
			if (ATH_TXQ_SETUP(sc, i))
				ath_tx_stopdma(sc, &sc->sc_txq[i]);
	}
	for (i = 0; i < HAL_NUM_TX_QUEUES; i++)
		if (ATH_TXQ_SETUP(sc, i))
			ath_tx_draintxq(sc, &sc->sc_txq[i]);
}

static struct sk_buff *
ath_alloc_skb(u_int size, u_int align)
{
	struct sk_buff *skb;
	u_int off;

	skb = dev_alloc_skb(size + align-1);
	if (skb != NULL) {
		off = ((unsigned long) skb->data) % align;
		if (off != 0)
			skb_reserve(skb, align - off);
	}
	return skb;
}

static int
find11gchannel(struct ieee80211com *ic, int i, int freq)
{
	for (; i < ic->ic_nchans; i++) {
		const struct ieee80211_channel *c = &ic->ic_channels[i];
		if (c->ic_freq == freq && IEEE80211_IS_CHAN_ANYG(c))
			return 1;
	}
	return 0;
}

static struct ieee80211_channel *
findchannel(struct ieee80211com *ic, int freq, int mode)
{
	static const u_int chanflags[] = {
		0,			/* IEEE80211_MODE_AUTO */
		IEEE80211_CHAN_A,	/* IEEE80211_MODE_11A */
		IEEE80211_CHAN_B,	/* IEEE80211_MODE_11B */
		IEEE80211_CHAN_PUREG,	/* IEEE80211_MODE_11G */
		IEEE80211_CHAN_FHSS,	/* IEEE80211_MODE_FH */
		IEEE80211_CHAN_108A,	/* IEEE80211_MODE_TURBO_A */
		IEEE80211_CHAN_108G,	/* IEEE80211_MODE_TURBO_G */
		IEEE80211_CHAN_ST,	/* IEEE80211_MODE_TURBO_STATIC_A */
	};
	u_int modeflags;
	int i;

	modeflags = chanflags[mode];
	for (i = 0; i < ic->ic_nchans; i++) {
		struct ieee80211_channel *c = &ic->ic_channels[i];

		if (c->ic_freq != freq)
			continue;
		if (mode == IEEE80211_MODE_AUTO) {
			/* ignore turbo channels for autoselect */
			if (IEEE80211_IS_CHAN_TURBO(c))
				continue;
			/*
			 * XXX special-case 11b/g channels so we
			 *     always select the g channel if both
			 *     are present.
			 */
			if (!IEEE80211_IS_CHAN_B(c) ||
			    !find11gchannel(ic, i+1, c->ic_freq))
				return c;
		} else {
			if ((c->ic_flags & modeflags) == modeflags)
				return c;
		}
	}
	return NULL;
}

static int
tx99_setup(struct ath_tx99_priv *tx99, u_int32_t *pRateCode)
{
	struct net_device *dev = tx99->dev;
	struct ath_softc *sc = dev->priv;
	struct ieee80211com *ic = &sc->sc_ic;
	const HAL_RATE_TABLE *rt;
	int i;

	printk ("in %s\n",__func__);
	/*
	 * Locate the channel to use if user-specified; otherwise
	 * user the current channel.
	 */
	if (tx99->txfreq != 0) {
		struct ieee80211_channel *c =
			findchannel(ic, tx99->txfreq, tx99->txmode);
		if (c == NULL) {
			printk("%s: channel %u mode %u not available\n",
				__func__, tx99->txfreq, tx99->txmode);
			return 0;
		}
		printk("%s: switching to channel %u (flags 0x%x mode %u)\n",
			__func__, c->ic_freq, c->ic_flags, tx99->txmode);
		ic->ic_curchan = c;
		ic->ic_set_channel(ic);
	}
	rt = sc->sc_currates;
	if (rt == NULL) {
		printk("%s: no rate table\n", __func__);
		return 0;
	}
	if (tx99->txrate != 0) {
		for (i = 0; i < rt->rateCount; i++) {
			if (tx99->txrate == rt->info[i].rateKbps)
				break;
		}
		if (i == rt->rateCount) {
			/*
			 * Requested rate not found; use the highest rate.
			 */
			i = rt->rateCount-1;
			printk("%s: txrate %u not found, using %u\n",
				__func__, tx99->txrate,
				rt->info[i].rateKbps);
		}
	} else {
		/*
		 * Use the lowest rate for the current channel.
		 */
		i = 0;
	}
	*pRateCode = rt->info[i].rateCode;

	printk("%s: using %u Kbps txrate (code 0x%x)\n",
		__func__, rt->info[i].rateKbps, rt->info[i].rateCode);
	return 1;
}

/*
 * Start continuous transmit operation.
 */
static void
tx99_start(struct ath_tx99 *arg)
{
#define	MIN(a,b)	((a) < (b) ? (a) : (b))
	struct ath_tx99_priv *tx99 = (struct ath_tx99_priv *) arg;
	struct net_device *dev = tx99->dev;
	struct ath_softc *sc = dev->priv;
	struct ath_hal *ah = sc->sc_ah;
	struct ieee80211com *ic = &sc->sc_ic;
	u_int8_t *p;
	u_int32_t txrate, r, i;
	u_int16_t hdrlen, framelen, dmalen, delay, totdelay;
	struct ieee80211_frame *hdr;
	struct ath_desc *ds;
	struct ath_buf *bf, *prev, *first, *last;
	struct ath_txq *txq;
	struct sk_buff *skb;
	int txantenna;
	u_int32_t val,xpaahigh, xpabhigh;

	printk("%s: device %s tx99 skb %p\n", __func__, dev->name, tx99->skb);

	if (tx99->skb != NULL) {		/* already active */
		printk("%s: already running\n", __func__);
		return;
	}
	tx99->skb = ath_alloc_skb(4096, 32);
	if (tx99->skb == NULL) {
		printk("%s: unable to allocate skb\n", __func__);
		return;
	}
	skb = tx99->skb;
	/*
	 * Setup channel using configured frequency+flags.
	 */
	if (!tx99_setup(tx99, &txrate)) {
		printk("%s: unable to setup operation\n", __func__);
		dev_kfree_skb(skb);
		tx99->skb = NULL;
		return;
	}

	/* build output packet */
	hdr = (struct ieee80211_frame *)skb_put(skb, sizeof(*hdr));
	IEEE80211_ADDR_COPY(&hdr->i_addr1, test_addr);
	IEEE80211_ADDR_COPY(&hdr->i_addr2, test_addr);
	IEEE80211_ADDR_COPY(&hdr->i_addr3, test_addr);
	hdr->i_dur[0] = 0x0;
	hdr->i_dur[1] = 0x0;
	hdr->i_seq[0] = 0x5a;
	hdr->i_seq[1] = 0x5a;
	hdr->i_fc[0] = IEEE80211_FC0_TYPE_DATA;
	hdr->i_fc[1] = 0;
	hdrlen = sizeof(*hdr);
	switch (tx99->mode) {
	case TX99_M_CONT_FRAME_DATA:
		for (r = 0; r < 2000; r += sizeof(PN9Data)) {
			p = skb_put(skb, sizeof(PN9Data));
			memcpy(p, PN9Data, sizeof(PN9Data));
		}
		break;
	case TX99_M_SINGLE_CARRIER:
		for (r = 0; r < 2000; r += sizeof(u_int32_t)) {
			u_int32_t z=0;
			p = skb_put(skb, sizeof(u_int32_t));
			memcpy(p, &z, sizeof(u_int32_t));
		}
		/* XXX Please change this to a single memset! */
		break;
	}

	txantenna = sc->sc_txantenna;
	if (txantenna == 0)		/* no auto, default to omni 1 */
		txantenna = 1;

	/* For single carreir mode, switch antenna to opposite of what we expect */
	if (tx99->mode == TX99_M_SINGLE_CARRIER) {
		txantenna = ((txantenna == 1) ? 2 : 1);
	}

	txq = sc->sc_ac2q[WME_AC_VO];
	printk("%s: using tx queue %d\n", __func__, txq->axq_qnum);

	ath_draintxq(sc);
	printk("%s: txpower set to %d dBm\n", __func__, tx99->txpower/2);
	/* send 20 frames for the Power Amp to settle down */
	prev = NULL;

	for (i = 0; i < 20; i++) {
		ATH_TXBUF_LOCK_BH(sc);
		bf = STAILQ_FIRST(&sc->sc_txbuf);
		if (bf != NULL)
			STAILQ_REMOVE_HEAD(&sc->sc_txbuf,bf_list);
		ATH_TXBUF_UNLOCK_BH(sc);				

		if (bf == NULL) {
			printk("%s: no tx buf\n", __func__);
			return;
		}
		dmalen = skb->len;
		framelen = dmalen + IEEE80211_CRC_LEN;

		bf->bf_skbaddr = bus_map_single(sc->sc_bdev,
			skb->data, dmalen, BUS_DMA_TODEVICE);
		bf->bf_skb = skb;
		bf->bf_node = NULL;
		ds = bf->bf_desc;
		r = ath_hal_setuptxdesc(ah, ds, framelen, hdrlen,
			HAL_PKT_TYPE_NORMAL,
			tx99->txpower,
			txrate,			/* tx rate */
			1,			/* max retries */
			HAL_TXKEYIX_INVALID,	/* no WEP */
			txantenna,		/* Omni Antenna 0 */
			HAL_TXDESC_CLRDMASK,
			0,			/* rts/cts rate */
			0,			/* rts/cts duration */
			0, 0, 0
		);
		if (r == AH_FALSE) {
			printk("%s: fail setuptxdesc r(%d)\n", __func__, r);
			return;
		}  

		r = ath_hal_filltxdesc(ah, ds, dmalen, AH_TRUE, AH_TRUE, ds);
		if (r == AH_FALSE) {
			printk("%s: fail fill tx desc r(%d)\n", __func__, r);
			return;
		}
		ath_hal_setupxtxdesc(ah, ds
		     , txrate, 15	/* series 1 */
		     , txrate, 15	/* series 2 */
		     , txrate, 15	/* series 3 */
		);
		if (prev != NULL)
			prev->bf_desc->ds_link = bf->bf_daddr;
		ds->ds_link = 0;
		ds->ds_data = bf->bf_skbaddr;
		/* insert the buffers in to tmp_q */
		STAILQ_INSERT_TAIL(&tx99->tmp_q, bf, bf_list);

		prev = bf;
	}

	first = STAILQ_FIRST(&tx99->tmp_q);
	last = STAILQ_LAST(&tx99->tmp_q, ath_buf, bf_list);

	ath_hal_intrset(ah, 0);    	/* disable interrupts */

	/* Force receive off */
	/* XXX re-enable RX in DIAG_SW as otherwise xmits hang below */
	ath_hal_stoppcurecv(ah);	/* disable PCU */
	OS_REG_CLR_BIT(ah, AR_DIAG_SW, AR_DIAG_RX_DIS);
	/*	OS_REG_SET_BIT(ah, AR_DIAG_SW, AR_DIAG_CHAN_IDLE);  */
	ath_hal_setrxfilter(ah, 0);	/* clear recv filter */
	ath_hal_stopdmarecv(ah);	/* disable DMA engine */
	mdelay(3);			/* 3ms is long enough for 1 frame */
	sc->sc_rxlink = NULL;		/* just in case */

	if (tx99->mode == TX99_M_SINGLE_CARRIER) {
		printk ("Single carrier mode\n");
		xpaahigh = OS_REG_READ(ah, AR_PHY_BB_XP_PA_CTL) & AR_PHY_BB_XPAA_ACTIVE_HIGH;
		val = OS_REG_READ(ah, AR_PHY_BB_XP_PA_CTL);
		xpabhigh = MS(val, AR_PHY_BB_XPAB_ACTIVE_HIGH);

		if (IEEE80211_IS_CHAN_A(ic->ic_curchan)) {
			val = OS_REG_READ(ah, AR_PHY_BB_XP_PA_CTL) & ~AR_PHY_BB_XPAA_ACTIVE_HIGH;
			val |= (!xpaahigh);
			OS_REG_WRITE(ah, AR_PHY_BB_XP_PA_CTL, val);
		} else {
			val = OS_REG_READ(ah, AR_PHY_BB_XP_PA_CTL) & ~AR_PHY_BB_XPAB_ACTIVE_HIGH;
			val |= SM(!xpabhigh, AR_PHY_BB_XPAB_ACTIVE_HIGH);
		}
	}

	ath_hal_puttxbuf(ah, txq->axq_qnum, first->bf_daddr);
	if (!ath_hal_txstart(ah, txq->axq_qnum))
		printk ("%s: txstart fialed, siabled by dfs?\n",__func__);

	totdelay = 0;
	STAILQ_FOREACH(bf, &tx99->tmp_q, bf_list) {
		for (delay = 0; ath_hal_txprocdesc(ah, bf->bf_desc) == HAL_EINPROGRESS && delay < 10; delay += 2)
			mdelay(2);
		if (delay > 10*1000) { /* 10 sec dead man timer */
			printk ("frame xmit did not complete in 10 secs, aborting\n");
			break;
		}
		totdelay += delay;
		printk (".");
	}
	printk (" %d msec to prime the PA (20 frames)\n", totdelay);

	/* sleep 20 ms (why?) */
	mdelay(20);

	/* Disable AGC to A2 */
	OS_REG_SET_BIT(ah, AR_PHY_TEST, PHY_AGC_CLR);

	OS_REG_SET_BIT(ah, 0x9864, 0x7f000);
	OS_REG_SET_BIT(ah, 0x9924, 0x7f00fe);
	OS_REG_SET_BIT(ah, AR_DIAG_SW, AR_DIAG_RX_CLR_HI | AR_DIAG_IGNORE_CS);

	OS_REG_WRITE(ah, AR_CR, AR_CR_RXD);	/* set receive disable */

	OS_REG_WRITE(ah, AR_DLCL_IFS(txq->axq_qnum), 0);

	OS_REG_WRITE(ah, AR_D_GBL_IFS_SIFS, 100);
	OS_REG_WRITE(ah, AR_D_GBL_IFS_EIFS, 100);
	OS_REG_WRITE(ah, AR_TIME_OUT, 0x400);

	/* enable prefetch on qnum */
	OS_REG_WRITE(ah, AR_D_FPCTL, 0x10|txq->axq_qnum);
	OS_REG_WRITE(ah, AR_DRETRY_LIMIT(txq->axq_qnum), 0xffffffff);

	/* set QCU modes */
	val = OS_REG_READ(ah, AR_QMISC(txq->axq_qnum));
	OS_REG_WRITE(ah, AR_QMISC(txq->axq_qnum),
		     val | AR_Q_MISC_DCU_EARLY_TERM_REQ);

	if (tx99->mode == TX99_M_SINGLE_CARRIER) {
		/* For Single carrier mode, we need to set a few more registers */
		/* set constant values in dac */
		OS_REG_WRITE(ah, AR_PHY_TSTDAC_CONST, SM(0x1FF, AR_PHY_TSTDAC_CONST_Q) |
			     AR_PHY_TSTDAC_CONST_I);
		/* Force TSTDAC with value in Reg 15 */
		val = OS_REG_READ(ah, AR_PHY_TESTCTRL);
		val &= 0xffffff7d;
		OS_REG_WRITE(ah, AR_PHY_TESTCTRL, val | AR_PHY_TESTCTRL_TXSRC_ALT
			     | AR_PHY_TESTCTRL_TXSRC_SRC);
		/* Manually turn on DAC */
		OS_REG_WRITE(ah, AR_PHY_ADC_CTL, 0x80038ffc);
	}
		
#ifdef TX99_EXPERIMENTAL
	/* NB: this is experimental code */
	if (tx99->mode == 3 || tx99->mode == 4) {
		u_int32_t txcfg, qbits;

		printk("%s: using experimental mode %u\n",
			__func__, tx99->mode);
		if (tx99->mode == 3) {
			OS_REG_WRITE(ah, AR_DLCL_IFS(txq->axq_qnum), 0);
			OS_REG_WRITE(ah, AR_DRETRY_LIMIT(txq->axq_qnum), 0xffffffff);
			OS_REG_WRITE(ah, AR_D_GBL_IFS_SIFS, 100);
			OS_REG_WRITE(ah, AR_D_GBL_IFS_EIFS, 100);
			OS_REG_WRITE(ah, AR_TIME_OUT, 0x400);
			OS_REG_WRITE(ah, AR_D_GBL_IFS_SLOT, 100);
		}
		OS_REG_WRITE(ah, AR_DRETRY_LIMIT(txq->axq_qnum), 0xffffffff);
		/* enable prefetch on qnum */
		OS_REG_WRITE(ah, AR_D_FPCTL, 0x10|txq->axq_qnum);
		txcfg = 5 | (6<<AR_FTRIG_S);
		OS_REG_WRITE(ah, AR_TXCFG, txcfg);

		/* set QCU modes */
		OS_REG_WRITE(ah, AR_QMISC(txq->axq_qnum),
			 AR_Q_MISC_DCU_EARLY_TERM_REQ
			 +AR_Q_MISC_FSP_ASAP
			 +AR_Q_MISC_CBR_INCR_DIS1
			 +AR_Q_MISC_CBR_INCR_DIS0
		);

		/* stop tx dma on other queues */
		qbits = 0x3ff;
		qbits &= ~(1<<txq->axq_qnum);
		for (i = 0; i < 10; i++) {
			if (i == txq->axq_qnum)
				continue;
			OS_REG_WRITE(ah, AR_Q_TXD, 1<<i);
		}
		OS_REG_WRITE(ah, AR_Q_TXD, qbits);

		/* clear and freeze MIB counters */
		OS_REG_WRITE(ah, AR_MIBC, AR_MIBC_CMC);
		OS_REG_WRITE(ah, AR_MIBC, AR_MIBC_FMC);
	 
		OS_REG_WRITE(ah, AR_DMISC(txq->axq_qnum),
			 (AR_D_MISC_ARB_LOCKOUT_CNTRL_GLOBAL <<
				AR_D_MISC_ARB_LOCKOUT_CNTRL_S)
			 +(AR_D_MISC_ARB_LOCKOUT_IGNORE)
			 +(AR_D_MISC_POST_FR_BKOFF_DIS)
			 +(AR_D_MISC_VIRT_COLL_POLICY)
			 +(AR_D_MISC_VIR_COL_HANDLING_IGNORE)
		);

		/* disconnect other QCUs */
		for (i = 0; i < HAL_NUM_TX_QUEUES+2; i++) {
			if (i == txq->axq_qnum)
				continue;
			OS_REG_WRITE(ah, AR_DQCUMASK(i), 0);
		}
	}
#endif /* TX99_EXPERIMENTAL */

	if (first == NULL) {
		printk ("First is null\n");
	}
	/*
	 * Start continuous transmit.
	 */
	last->bf_desc->ds_link = first->bf_daddr; /* link to self */

	ath_hal_puttxbuf(ah, txq->axq_qnum, first->bf_daddr);


	if (tx99->mode == TX99_M_SINGLE_CARRIER) {
		/* turn on rf */
		val = 0x10a098c2;
		if (!IEEE80211_IS_CHAN_A(ic->ic_curchan))
			val |= 0x400000;
		OS_REG_WRITE(ah, 0x98DC, val);
		mdelay(10);
		val |= 0x4000;
		OS_REG_WRITE(ah, 0x98DC, val);
	} else {
		ath_hal_txstart(ah, txq->axq_qnum);
	}

	printk("%s: continuous transmit started\n", __func__);
}

/*
 * Terminate continuous transmit operation.
 */
static void
tx99_stop(struct ath_tx99 *arg)
{
	struct ath_tx99_priv *tx99 = (struct ath_tx99_priv *) arg;
	struct net_device *dev = tx99->dev;
	struct ath_softc *sc = dev->priv;
	struct ath_hal *ah = sc->sc_ah;
	struct ieee80211com *ic = &sc->sc_ic;
	struct ath_buf *bf;
	struct sk_buff *skb;
	int j;

	printk("%s: device %s skb %p\n", __func__, dev->name, tx99->skb);

	if (tx99->skb == NULL)
		return;
	skb = tx99->skb;
	tx99->skb = NULL;

	OS_REG_WRITE(ah, AR_PHY_TEST,
		OS_REG_READ(ah, AR_PHY_TEST) &~ PHY_AGC_CLR);
	OS_REG_WRITE(ah, AR_DIAG_SW,
		OS_REG_READ(ah, AR_DIAG_SW) &~ (AR_DIAG_RX_CLR_HI | AR_DIAG_IGNORE_CS));
	/* return the buffers to the txbuf list */

	ATH_TXBUF_LOCK_BH(sc);						
	for (j = 0; (bf = STAILQ_FIRST(&tx99->tmp_q)) != NULL; j++) {
		bf->bf_skb = NULL;
		STAILQ_REMOVE_HEAD(&tx99->tmp_q, bf_list);
		STAILQ_INSERT_HEAD(&sc->sc_txbuf, bf, bf_list);
	}
	ATH_TXBUF_UNLOCK_BH(sc);				
	printk("%s: reclaimed %d buffers \n", __func__, j);

	ic->ic_reset(ic->ic_dev);
	dev_kfree_skb(skb);

	printk("%s: continuous transmit stopped\n", __func__);
}

enum {
	ATH_MODE	= 100,		/* enable experimental modes */
	ATH_TXRATE	= 101,		/* xmit rate */
	ATH_TXPOWER	= 102,		/* xmit power */
	ATH_TXFREQ	= 103,		/* channel freq */
	ATH_TXMODE	= 104,		/* band/mode, e.g. 11b */
};

static int
ATH_SYSCTL_DECL(ath_sysctl_tx99param, ctl, write, filp, buffer, lenp, ppos)
{
	struct ath_tx99_priv *tx99 = ctl->extra1;
	u_int val;
	int ret;

	ctl->data = &val;
	ctl->maxlen = sizeof(val);
	if (write) {
		ret = ATH_SYSCTL_PROC_DOINTVEC(ctl, write, filp, buffer,
				lenp, ppos);
		if (ret == 0) {
			switch (ctl->ctl_name) {
			case ATH_MODE:
				tx99->mode = val;
				break;
			case ATH_TXPOWER:
				tx99->txpower = val*2;
				break;
			case ATH_TXRATE:
				tx99->txrate = val;
				break;
			case ATH_TXMODE:
				tx99->txmode = val;
				break;
			case ATH_TXFREQ:
				tx99->txfreq = val;
				break;
			default:
				return -EINVAL;
			}
		}
	} else {
		switch (ctl->ctl_name) {
		case ATH_MODE:
			val = tx99->mode;
			break;
		case ATH_TXPOWER:
			val = tx99->txpower/2;
			break;
		case ATH_TXRATE:
			val = tx99->txrate;
			break;
		case ATH_TXMODE:
			val = tx99->txmode;
			break;
		case ATH_TXFREQ:
			val = tx99->txfreq;
			break;
		}
		ret = ATH_SYSCTL_PROC_DOINTVEC(ctl, write, filp, buffer,
				lenp, ppos);
	}
	return ret;
}

#define	CTL_AUTO	-2	/* cannot be CTL_ANY or CTL_NONE */

static const ctl_table ath_sysctl_template[] = {
	{ .ctl_name	= ATH_MODE,
	  .procname	= "tx99config",
	  .mode		= 0644,
	  .proc_handler	= ath_sysctl_tx99param
	},
	{ .ctl_name	= ATH_TXRATE,
	  .procname	= "tx99rate",
	  .mode		= 0644,
	  .proc_handler	= ath_sysctl_tx99param
	},
	{ .ctl_name	= ATH_TXPOWER,
	  .procname	= "tx99power",
	  .mode		= 0644,
	  .proc_handler	= ath_sysctl_tx99param
	},
	{ .ctl_name	= ATH_TXFREQ,
	  .procname	= "tx99freq",
	  .mode		= 0644,
	  .proc_handler	= ath_sysctl_tx99param
	},
	{ .ctl_name	= ATH_TXMODE,
	  .procname	= "tx99mode",
	  .mode		= 0644,
	  .proc_handler	= ath_sysctl_tx99param
	},
	{ 0 }
};

static void
tx99_sysctl_register(struct ath_tx99_priv *tx99)
{
	int i, space;

	space = 5*sizeof(struct ctl_table) + sizeof(ath_sysctl_template);
	tx99->sysctls = kmalloc(space, GFP_KERNEL);
	if (tx99->sysctls == NULL) {
		printk("%s: no memory for sysctl table!\n", __func__);
		return;
	}

	/* setup the table */
	memset(tx99->sysctls, 0, space);
	tx99->sysctls[0].ctl_name = CTL_DEV;
	tx99->sysctls[0].procname = "dev";
	tx99->sysctls[0].mode = 0555;
	tx99->sysctls[0].child = &tx99->sysctls[2];
	/* [1] is NULL terminator */
	tx99->sysctls[2].ctl_name = CTL_AUTO;
	tx99->sysctls[2].procname = tx99->dev->name;
	tx99->sysctls[2].mode = 0555;
	tx99->sysctls[2].child = &tx99->sysctls[4];
	/* [3] is NULL terminator */
	/* copy in pre-defined data */
	memcpy(&tx99->sysctls[4], ath_sysctl_template,
		sizeof(ath_sysctl_template));

	/* add in dynamic data references */
	for (i = 4; tx99->sysctls[i].ctl_name; i++)
		if (tx99->sysctls[i].extra1 == NULL)
			tx99->sysctls[i].extra1 = tx99;

	/* and register everything */
	tx99->sysctl_header = register_sysctl_table(tx99->sysctls, 1);
	if (!tx99->sysctl_header) {
		printk("%s: failed to register tx99 sysctls!\n",
			tx99->dev->name);
		kfree(tx99->sysctls), tx99->sysctls = NULL;
	}
}

static void
tx99_detach(struct ath_tx99 *arg)
{
	struct ath_tx99_priv *tx99 = (struct ath_tx99_priv *) arg;
	struct net_device *dev = tx99->dev;
	struct ath_softc *sc = dev->priv;

 	printk("%s: device %s\n", __func__, dev->name);

	sc->sc_tx99 = NULL;
	tx99_stop(arg);
	if (tx99->sysctl_header != NULL)
		unregister_sysctl_table(tx99->sysctl_header);
	if (tx99->sysctls != NULL)
		kfree(tx99->sysctls);
	kfree(tx99);
}

static void
tx99_attach(struct net_device *dev)
{
	struct ath_softc *sc = dev->priv;
	struct ath_tx99_priv *tx99;

	printk("%s: device %s\n", __func__, dev->name);

	tx99 = kmalloc(sizeof(struct ath_tx99_priv), GFP_KERNEL);
	if (tx99 == NULL) {
		printk(KERN_INFO "%s: no memory for tx99 state\n", __func__);
		return;
	}
	memset(tx99, 0, sizeof(*tx99));
	tx99->base.stop = tx99_stop;
	tx99->base.start = tx99_start;
	tx99->base.detach = tx99_detach;
	tx99->dev = dev;
	STAILQ_INIT(&tx99->tmp_q);
	tx99->txpower = 60;
	tx99->txfreq = 0;
	tx99->txmode = IEEE80211_MODE_AUTO;
	tx99_sysctl_register(tx99);
	sc->sc_tx99 = &tx99->base;
}

/*
 * Module glue.
 */
static	char *dev_info = "ath_tx99";

static char *device = NULL;		/* ath device to use for testing */
static struct net_device *tx99_dev = NULL;

MODULE_AUTHOR("Atheros");
MODULE_DESCRIPTION("tx99 support for Atheros 5212 devices");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Proprietary");
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(device, "s");
#else
module_param(device, charp, 0444);
#endif


static int __init
init_ath_tx99(void)
{	
    /* XXX version is a guess; where should it come from? */
	printk(KERN_INFO "%s: Version 1.0\n"
		"Copyright (c) 2004-2005 Atheros Communications, Inc, "
		"All Rights Reserved\n", dev_info);
	if (device == NULL) {
		printk(KERN_INFO "%s: no device specified for testing\n",
			dev_info);
		return -ENOSYS;
	}
	tx99_dev = dev_get_by_name(device);
	if (tx99_dev == NULL) {
		printk(KERN_INFO "%s: device %s not found\n", dev_info, device);
		return -ENOSYS;
	}
	tx99_attach(tx99_dev);
	return 0;
}
module_init(init_ath_tx99);

static void __exit
exit_ath_tx99(void)
{
	if (tx99_dev != NULL) {
		struct ath_softc *sc = tx99_dev->priv;
		if (sc->sc_tx99 != NULL)
			tx99_detach(sc->sc_tx99);
		dev_put(tx99_dev);		/* release ref */
	}
	printk(KERN_INFO "%s: driver unloaded\n", dev_info);
}
module_exit(exit_ath_tx99);
